class HeimdallrClientException(Exception):
    """ Basic Heimdallr client exception.

    A HeimdallrClientException is raised when an error is received
    from the Heimdallr server.
    """
    pass